/* scherzo - Music notation training
 *
 * Copyright © 2010 Carl Worth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/ .
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <asoundlib.h>

#include "score.h"
#include "mnemon.h"

#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof(arr[0]))

#define unused(foo) foo __attribute__((unused))

#define MIDI_BUF_SIZE 4096

typedef struct challenge
{
    int active;
    bin_t *bin;
    int item_index;
    score_staff_t *staff;
    pitch_t pitch;

    int satisfied;
    int mistaken;
} challenge_t;

typedef struct pitch_group
{
    void *ctx;
    pitch_t *pitches;
    int size;
    int num_pitches;
} pitch_group_t;

typedef struct scherzo
{
    void *ctx;

    GtkWidget *window;
    score_t *score;
    int staff_height;

    score_staff_t *treble;
    score_staff_t *bass;
    score_chord_t *chord;

    /* The word "keyboard" here is referring to a "computer
     * keyboard". Any "piano keyboard" key knows its own octave and
     * accidental already. */
    int keyboard_octave;
    pitch_accidental_t keyboard_accidental;

    int midi_fd;
    snd_midi_event_t *snd_midi_event;

    mnemon_t mnemon;
    challenge_t challenge;

    pitch_group_t notes_pressed;
    pitch_group_t notes_pedaled;

    int pedal_pressed;
} scherzo_t;

/* Forward declarations. */
static void
scherzo_press_note (scherzo_t *scherzo, pitch_t pitch);

static void
scherzo_release_note (scherzo_t *scherzo, pitch_t pitch);

static void
scherzo_press_pedal (scherzo_t *scherzo);

static void
scherzo_release_pedal (scherzo_t *scherzo);

static void
_judge_pitch (scherzo_t *scherzo, pitch_t pitch);

static void
_score_challenge (scherzo_t *scherzo);

static int
on_delete_event_quit (unused (GtkWidget *widget),
		      unused (GdkEvent *event),
		      unused (gpointer user_data))
{
    gtk_main_quit ();

    /* Returning FALSE allows the default handler for delete-event
     * to proceed to cleanup the widget. */
    return FALSE;
}

static int
on_expose_event_draw (GtkWidget *widget,
		      unused (GdkEventExpose *expose),
		      void * user_data)
{
    scherzo_t *scherzo = user_data;
    score_t *score = scherzo->score;
    cairo_t *cr;
    GtkAllocation allocation;
    static const int pad = 10;
    int widget_width;

    gtk_widget_get_allocation (widget, &allocation);
    widget_width = allocation.width;

    cr = gdk_cairo_create (widget->window);

    /* White background */
    cairo_set_source_rgb (cr, 1.0, 1.0, 1.0);
    cairo_paint (cr);

    /* Add some padding on the sides and top */
    cairo_translate (cr, pad, scherzo->staff_height);
    score_set_staff_height (score, scherzo->staff_height);
    score_set_width (score, widget_width - 2 * pad);

    score_draw (score, cr);
 
    return TRUE;
}

static int
on_key_press_event (GtkWidget *widget,
		    GdkEventKey *key,
		    void *user_data)
{
    scherzo_t *scherzo = user_data;
    int octave;
    /* Initialize to keep the compiler quiet. */
    pitch_name_t pitch_name = PITCH_NATURAL (C, 0);
    pitch_t pitch;

    if (scherzo->challenge.active)
	octave = PITCH_OCTAVE (scherzo->challenge.pitch);
    else
	octave = scherzo->keyboard_octave;

    switch (key->keyval) {
    case GDK_KEY_plus:
    case GDK_KEY_KP_Add:
    case GDK_KEY_equal:
    case GDK_KEY_KP_Equal:
	scherzo->staff_height += 4;
	gtk_widget_queue_draw (widget);
	return TRUE;
	break;
    case GDK_KEY_minus:
    case GDK_KEY_KP_Subtract:
	scherzo->staff_height -= 4;
	gtk_widget_queue_draw (widget);
	return TRUE;
	break;
    case GDK_KEY_q:
    case GDK_KEY_Q:
    case GDK_KEY_Escape:
	gtk_main_quit ();
	return FALSE;
    case GDK_KEY_c:
    case GDK_KEY_C:
	pitch_name = PITCH_NAME_C;
	break;
    case GDK_KEY_d:
    case GDK_KEY_D:
	pitch_name = PITCH_NAME_D;
	break;
    case GDK_KEY_e:
    case GDK_KEY_E:
	pitch_name = PITCH_NAME_E;
	break;
    case GDK_KEY_f:
    case GDK_KEY_F:
	pitch_name = PITCH_NAME_F;
	break;
    case GDK_KEY_g:
    case GDK_KEY_G:
	pitch_name = PITCH_NAME_G;
	break;
    case GDK_KEY_a:
    case GDK_KEY_A:
	pitch_name = PITCH_NAME_A;
	break;
    case GDK_KEY_b:
    case GDK_KEY_B:
	pitch_name = PITCH_NAME_B;
	break;
    case GDK_KEY_0:
    case GDK_KEY_1:
    case GDK_KEY_2:
    case GDK_KEY_3:
    case GDK_KEY_4:
    case GDK_KEY_5:
    case GDK_KEY_6:
    case GDK_KEY_7:
    case GDK_KEY_8:
	scherzo->keyboard_octave = key->keyval - GDK_KEY_0;
	break;
    case GDK_KEY_space:
	scherzo_press_pedal (scherzo);
	break;
    case GDK_KEY_Up:
	if (scherzo->keyboard_accidental < PITCH_ACCIDENTAL_DOUBLE_SHARP)
	    scherzo->keyboard_accidental++;
	break;
    case GDK_KEY_Down:
	if (scherzo->keyboard_accidental > PITCH_ACCIDENTAL_DOUBLE_FLAT)
	    scherzo->keyboard_accidental--;
	break;
    }

    pitch = PITCH (pitch_name, scherzo->keyboard_accidental, octave);

    if ((key->keyval >= GDK_KEY_A && key->keyval <= GDK_KEY_G) ||
	(key->keyval >= GDK_KEY_a && key->keyval <= GDK_KEY_g))
    {
	scherzo_press_note (scherzo, pitch);
	_judge_pitch (scherzo, pitch);
	gtk_widget_queue_draw (scherzo->window);

	return TRUE;
    }


    /* Allow an unhandled event to propagate to other handlers. */
    return FALSE;
}

static int
on_key_release_event (unused (GtkWidget *widget),
		      GdkEventKey *key,
		      void *user_data)
{
    scherzo_t *scherzo = user_data;
    int octave;
    /* Initialize to keep the compiler quiet. */
    pitch_name_t pitch_name = PITCH_NAME_C;
    pitch_t pitch;

    if (scherzo->challenge.active)
	octave = PITCH_OCTAVE (scherzo->challenge.pitch);
    else
	octave = scherzo->keyboard_octave;

    switch (key->keyval) {
    case GDK_KEY_c:
    case GDK_KEY_C:
	pitch_name = PITCH_NAME_C;
	break;
    case GDK_KEY_d:
    case GDK_KEY_D:
	pitch_name = PITCH_NAME_D;
	break;
    case GDK_KEY_e:
    case GDK_KEY_E:
	pitch_name = PITCH_NAME_E;
	break;
    case GDK_KEY_f:
    case GDK_KEY_F:
	pitch_name = PITCH_NAME_F;
	break;
    case GDK_KEY_g:
    case GDK_KEY_G:
	pitch_name = PITCH_NAME_G;
	break;
    case GDK_KEY_a:
    case GDK_KEY_A:
	pitch_name = PITCH_NAME_A;
	break;
    case GDK_KEY_b:
    case GDK_KEY_B:
	pitch_name = PITCH_NAME_B;
	break;
    case GDK_KEY_space:
	scherzo_release_pedal (scherzo);
	break;
    }

    pitch = PITCH (pitch_name, scherzo->keyboard_accidental, octave);

    if ((key->keyval >= GDK_KEY_A && key->keyval <= GDK_KEY_G) ||
	(key->keyval >= GDK_KEY_a && key->keyval <= GDK_KEY_g))
    {
	scherzo_release_note (scherzo, pitch);
	_score_challenge (scherzo);
	gtk_widget_queue_draw (scherzo->window);

	return TRUE;
    }


    /* Allow an unhandled event to propagate to other handlers. */
    return FALSE;
}

static unsigned char
_pitch_to_midi (pitch_t pitch)
{
    int octave = PITCH_OCTAVE (pitch);
    unsigned char midi_note = 12 * (octave + 1);

    switch (PITCH_NAME (pitch)) {
    case PITCH_NAME_C:
	break;
    case PITCH_NAME_D:
	midi_note += 2;
	break;
    case PITCH_NAME_E:
	midi_note += 4;
	break;
    case PITCH_NAME_F:
	midi_note += 5;
	break;
    case PITCH_NAME_G:
	midi_note += 7;
	break;
    case PITCH_NAME_A:
	midi_note += 9;
	break;
    case PITCH_NAME_B:
	midi_note += 11;
	break;
    }

    switch (PITCH_ACCIDENTAL (pitch)) {
    case PITCH_ACCIDENTAL_DOUBLE_FLAT:
	midi_note -= 2;
	break;
    case PITCH_ACCIDENTAL_FLAT:
	midi_note -= 1;
	break;
    case PITCH_ACCIDENTAL_NATURAL:
	break;
    case PITCH_ACCIDENTAL_SHARP:
	midi_note += 1;
	break;
    case PITCH_ACCIDENTAL_DOUBLE_SHARP:
	midi_note += 2;
	break;
    }

    return midi_note;
}

/* octave can optionally by NULL */
static pitch_t
_midi_to_pitch (unsigned char midi_note)
{
    int octave = octave = midi_note / 12 - 1;

    switch (midi_note % 12)
    {
    default:
    case 0:
	return PITCH_LITERAL (C, NATURAL, octave);
	break;
    case 1:
	return PITCH_LITERAL (C, SHARP, octave);
	break;
    case 2:
	return PITCH_LITERAL (D, NATURAL, octave);
	break;
    case 3:
	return PITCH_LITERAL (D, SHARP, octave);
	break;
    case 4:
	return PITCH_LITERAL (E, NATURAL, octave);
	break;
    case 5:
	return PITCH_LITERAL (F, NATURAL, octave);
	break;
    case 6:
	return PITCH_LITERAL (F, SHARP, octave);
	break;
    case 7:
	return PITCH_LITERAL (G, NATURAL, octave);
	break;
    case 8:
	return PITCH_LITERAL (G, SHARP, octave);
	break;
    case 9:
	return PITCH_LITERAL (A, NATURAL, octave);
	break;
    case 10:
	return PITCH_LITERAL (A, SHARP, octave);
	break;
    case 11:
	return PITCH_LITERAL (B, NATURAL, octave);
	break;
    }
}

/* Determine a chord name (if any) from the current notes pressed */

typedef struct analyzed_pitch {
    /* Original note being analzyed. */
    pitch_t pitch;

    /* Absolute pitch (expressed as midi number). */
    int midi_pitch;

    /* Pitch relative to bass note. */
    int relative_pitch;
} analyzed_pitch_t;

static int
_compare_analyzed_pitch_by_midi_pitch (const void *va, const void *vb)
{
    const analyzed_pitch_t *a = va, *b = vb;

    return a->midi_pitch - b->midi_pitch;
}

static int
_compare_analyzed_pitch_by_relative_pitch (const void *va, const void *vb)
{
    const analyzed_pitch_t *a = va, *b = vb;

    return a->relative_pitch - b->relative_pitch;
}

typedef struct modified_degree
{
    int degree;
    int modification;
} modified_degree_t;

static int
_modified_degree_to_half_steps (const modified_degree_t *degree)
{
    int half_steps;
    int scale_degree = degree->degree;

    /* Restrict to actual degrees within a scale. */
    if (scale_degree > 7)
	scale_degree = (scale_degree % 8) + 1;

    /* Number of half steps from root to specified degree within a
     * diatonic scale. */
    switch (scale_degree) {
    case 1:
	half_steps = 0;
	break;
    case 2:
	half_steps = 2;
	break;
    case 3:
	half_steps = 4;
	break;
    case 4:
	half_steps = 5;
	break;
    case 5:
	half_steps = 7;
	break;
    case 6:
	half_steps = 9;
	break;
    case 7:
	half_steps = 11;
	break;
    default:
	fprintf (stderr, "Internal: Invalid degree %d\n", degree->degree);
	exit (1);
	break;
    }

    return half_steps + degree->modification;
}

/* Number of half steps from 'root' up to 'pitch' (that is, counting
 * the steps up a chromatic scale), within the same octave. */
static int
_pitch_from_root_in_half_steps (pitch_t pitch, pitch_t root)
{
    int root_midi = _pitch_to_midi (root);
    int pitch_midi = _pitch_to_midi (pitch);

    while (pitch_midi < root_midi)
	pitch_midi += 12;

    return (pitch_midi - root_midi) % 12;
}

static int
_chord_signature_matches (analyzed_pitch_t *pitches,
			  int num_pitches,
			  modified_degree_t *degrees,
			  int num_degrees,
			  int inversion,
			  pitch_t *root)
{
#define MAX_DEGREES 6
    int relative_pitches[MAX_DEGREES];
    int i, root_index;

    assert (num_degrees <= MAX_DEGREES);

    if (num_pitches != num_degrees)
	return 0;

    if (inversion >= num_degrees)
	return 0;

    /* We never spell simple intervals as inversions. */
    if (num_degrees == 2 && inversion > 0)
	return 0;

    for (i = 0; i < num_degrees; i++) {
	/* The num_degrees is in the addition just to ensure all
	 * inputs to the modulus operator remain positive. */
	int index = (i + num_degrees - inversion) % num_degrees;

	/* Again, adding a 12 to keep things positive. */
	relative_pitches[index] =
	    (12 +
	     _modified_degree_to_half_steps (&degrees[i]) -
	     _modified_degree_to_half_steps (&degrees[inversion])) % 12;
		
    }

    for (i = 0; i < num_pitches; i++)
	if (pitches[i].relative_pitch != relative_pitches[i])
	    return 0;

    root_index = (num_pitches - inversion) % num_pitches;
    *root = pitches[root_index].pitch;

    return 1;
}

static const char *
_pitch_str (pitch_t pitch)
{
    static char double_flat[] = "X𝄫";
    static char flat[] = "X♭";
    static char natural[] = "X";
    static char sharp[] = "X♯";
    static char double_sharp[] = "X𝄪";
    char *ret;

    switch (PITCH_ACCIDENTAL (pitch)) {
    case PITCH_ACCIDENTAL_DOUBLE_FLAT:
	ret = double_flat;
	break;
    case PITCH_ACCIDENTAL_FLAT:
	ret = flat;
	break;
    case PITCH_ACCIDENTAL_NATURAL:
	ret = natural;
	break;
    case PITCH_ACCIDENTAL_SHARP:
	ret = sharp;
	break;
    case PITCH_ACCIDENTAL_DOUBLE_SHARP:
	ret = double_sharp;
	break;
    }

    switch (PITCH_NAME(pitch)) {
    case PITCH_NAME_C:
	ret[0] = 'C';
	break;
    case PITCH_NAME_D:
	ret[0] = 'D';
	break;
    case PITCH_NAME_E:
	ret[0] = 'E';
	break;
    case PITCH_NAME_F:
	ret[0] = 'F';
	break;
    case PITCH_NAME_G:
	ret[0] = 'G';
	break;
    case PITCH_NAME_A:
	ret[0] = 'A';
	break;
    case PITCH_NAME_B:
	ret[0] = 'B';
	break;
    }

    return ret;
}

typedef struct chord_signature
{
    int num_degrees;
    modified_degree_t degrees[MAX_DEGREES];
    const char *name;
} chord_signature_t;

/* XXX: The superscript rise value should be relative to the current
 * font size. Best would be for pango to support this. See:
 *
 * https://bugzilla.gnome.org/show_bug.cgi?id=708780
 *
 * For now, these are emipirically-derived values that look reasonable
 * at the default size that score.c is using to draw the staff. If the
 * user scales the staff up or down then these will look wrong.
 */

#define SUP "<span font='33' rise='30000'>"
#define PUS "</span>"

chord_signature_t signatures[] = {
    /* Octave */
    { 1, {{1, 0}}, "P8" },

    /* Intervals */
    { 2, {{1, 0}, {2, -1}}, "m2" },
    { 2, {{1, 0}, {2,  0}}, "M2" },
    { 2, {{1, 0}, {3, -1}}, "m3" },
    { 2, {{1, 0}, {3,  0}}, "M3" },
    { 2, {{1, 0}, {4,  0}}, "P4" },
    { 2, {{1, 0}, {5, -1}}, "Tri-tone" },
    { 2, {{1, 0}, {5,  0}}, "P5" },
    { 2, {{1, 0}, {6, -1}}, "m6" },
    { 2, {{1, 0}, {6,  0}}, "M6" },
    { 2, {{1, 0}, {7, -1}}, "m7" },
    { 2, {{1, 0}, {7,  0}}, "M7" },

    /* Triads */
    { 3, {{1, 0}, {4,  0}, {5,  0}}, "sus" },
    { 3, {{1, 0}, {3,  0}, {5, +1}}, SUP "+" PUS },
    { 3, {{1, 0}, {3,  0}, {5,  0}}, "" },
    { 3, {{1, 0}, {3, -1}, {5,  0}}, "m" },
    { 3, {{1, 0}, {3, -1}, {5, -1}}, "°" },
    { 3, {{1, 0}, {2,  0}, {5,  0}}, "msus2" },

    /* Seventh chords with no 3rd */
    { 3, {{1, 0}, {5, +1}, {7,  0}}, SUP "+M7" PUS },
    { 3, {{1, 0}, {5, +1}, {7, -1}}, SUP "+7" PUS },
    { 3, {{1, 0}, {5,  0}, {7,  0}}, "M7" },
    { 3, {{1, 0}, {5,  0}, {7, -1}}, "7" },
    { 3, {{1, 0}, {5, -1}, {7, -1}}, "7♭5" },

    /* Seventh chords with no 5th */
    { 3, {{1, 0}, {4,  0}, {7,  0}}, "M7sus" },
    { 3, {{1, 0}, {4,  0}, {7, -1}}, "7sus" },
    { 3, {{1, 0}, {3,  0}, {7,  0}}, "M7" },
    { 3, {{1, 0}, {3,  0}, {7, -1}}, "7" },
    { 3, {{1, 0}, {3, -1}, {7,  0}}, "m" SUP "M7" PUS },
    { 3, {{1, 0}, {3, -1}, {7, -1}}, "m7" },
    { 3, {{1, 0}, {2,  0}, {7,  0}}, "m" SUP "M7" PUS "sus2" },
    { 3, {{1, 0}, {2,  0}, {7, -1}}, "m7sus2" },

    /* Sixth chords */
    { 4, {{1, 0}, {4,  0}, {5,  0}, {6,  0}}, "6sus" },
    { 4, {{1, 0}, {3,  0}, {5,  0}, {6,  0}}, "6" }, /* Ambiguous with m7 */
    { 4, {{1, 0}, {3, -1}, {5,  0}, {6,  0}}, "m6" },
    { 4, {{1, 0}, {2,  0}, {5,  0}, {6,  0}}, "m6sus2" },

    /* Seventh chords */
    { 4, {{1, 0}, {4,  0}, {5, +1}, {7,  0}}, SUP "+M7" PUS "sus" },
    { 4, {{1, 0}, {4,  0}, {5, +1}, {7, -1}}, SUP "+7" PUS "sus" },
    { 4, {{1, 0}, {4,  0}, {5,  0}, {7,  0}}, "M7sus" },
    { 4, {{1, 0}, {4,  0}, {5,  0}, {7, -1}}, "7sus" },
    { 4, {{1, 0}, {4,  0}, {5, -1}, {7, -1}}, "7♭5sus" },
    { 4, {{1, 0}, {3,  0}, {5, +1}, {7,  0}}, SUP "+M7" PUS },
    { 4, {{1, 0}, {3,  0}, {5, +1}, {7, -1}}, SUP "+7" PUS },
    { 4, {{1, 0}, {3,  0}, {5,  0}, {7,  0}}, "M7" },
    { 4, {{1, 0}, {3,  0}, {5,  0}, {7, -1}}, "7" },
    { 4, {{1, 0}, {3,  0}, {5, -1}, {7, -1}}, "7♭5" },
    { 4, {{1, 0}, {3, -1}, {5,  0}, {7,  0}}, "m" SUP "M7" PUS },
    { 4, {{1, 0}, {3, -1}, {5,  0}, {7, -1}}, "m7" },
    { 4, {{1, 0}, {3, -1}, {5, -1}, {7,  0}}, "°" SUP "M7" PUS },
    { 4, {{1, 0}, {3, -1}, {5, -1}, {7, -1}}, "𝆩" SUP "7" PUS },
    { 4, {{1, 0}, {3, -1}, {5, -1}, {7, -2}}, "°" SUP "7" PUS },
    { 4, {{1, 0}, {2,  0}, {5,  0}, {7,  0}}, "m" SUP "M7" PUS "sus2" },
    { 4, {{1, 0}, {2,  0}, {5,  0}, {7, -1}}, "m7sus2" },
    { 4, {{1, 0}, {2,  0}, {5, -1}, {7,  0}}, "°" SUP "M7" PUS "sus2" },
    { 4, {{1, 0}, {2,  0}, {5, -1}, {7, -1}}, "𝆩" SUP "7" PUS "sus2" },
    { 4, {{1, 0}, {2,  0}, {5, -1}, {7, -2}}, "°" SUP "7" PUS "sus2" }, /* Ambiguous with 7 */

    /* The sorting for chords with degree higher than 9 is funny
     * to keep the array in degree order after reducing each
     * degree to an actual scale degree, (9 -> 2, 11 -> 4, 13 ->
     * 6) */

    /* Ninth chords voiced with no 5th */
    { 4, {{1, 0}, {9,  0}, {4,  0}, {7,  0}}, "M9sus" },
    { 4, {{1, 0}, {9,  0}, {4,  0}, {7, -1}}, "9sus" },
    { 4, {{1, 0}, {9,  0}, {3,  0}, {7,  0}}, "M9" },
    { 4, {{1, 0}, {9,  0}, {3,  0}, {7, -1}}, "9" },
    { 4, {{1, 0}, {9,  0}, {3, -1}, {7,  0}}, "m" SUP "M9" PUS },
    { 4, {{1, 0}, {9,  0}, {3, -1}, {7, -1}}, "m9" },
    { 4, {{1, 0}, {9, -1}, {3, -1}, {7, -1}}, "m♭9" },
    { 4, {{1, 0}, {9,  0}, {3, -1}, {6,  0}}, "m6/9" },
    { 4, {{1, 0}, {9, -1}, {3, -1}, {6,  0}}, "m6/♭9" },

    /* Sixth plus 9 */
    { 5, {{1, 0}, {9,  0}, {3,  0}, {5,  0}, {6,  0}}, "6/9" },
    { 5, {{1, 0}, {9,  0}, {3, -1}, {5,  0}, {6,  0}}, "m6/9" },
    { 5, {{1, 0}, {9, -1}, {3, -1}, {5,  0}, {6,  0}}, "m6/♭9" },

    /* Seventh plus altered 9 */
    { 5, {{1, 0}, {9, +1}, {3,  0}, {5,  0}, {7, -1}}, "7♯9" },
    { 5, {{1, 0}, {9, -1}, {3,  0}, {5,  0}, {7, -1}}, "7♭9" },
    { 5, {{1, 0}, {9, +1}, {3,  0}, {5, -1}, {7, -1}}, "7♭5♯9" },
    { 5, {{1, 0}, {9, -1}, {3,  0}, {5, -1}, {7, -1}}, "7♭5♭9" },

    /* Ninth chords */
    { 5, {{1, 0}, {9,  0}, {4,  0}, {5, +1}, {7,  0}}, SUP "+M9" PUS "sus" },
    { 5, {{1, 0}, {9,  0}, {4,  0}, {5, +1}, {7, -1}}, SUP "+9" PUS "sus" },
    { 5, {{1, 0}, {9,  0}, {4,  0}, {5,  0}, {7,  0}}, "M9sus" },
    { 5, {{1, 0}, {9,  0}, {4,  0}, {5,  0}, {7, -1}}, "9sus" },
    { 5, {{1, 0}, {9,  0}, {3,  0}, {5, +1}, {7,  0}}, SUP "+M9" PUS },
    { 5, {{1, 0}, {9,  0}, {3,  0}, {5, +1}, {7, -1}}, SUP "+9" PUS },
    { 5, {{1, 0}, {9,  0}, {3,  0}, {5,  0}, {7,  0}}, "M9" },
    { 5, {{1, 0}, {9,  0}, {3,  0}, {5,  0}, {7, -1}}, "9" },
    { 5, {{1, 0}, {9,  0}, {3, -1}, {5,  0}, {7,  0}}, "m" SUP "M9" PUS },
    { 5, {{1, 0}, {9,  0}, {3, -1}, {5,  0}, {7, -1}}, "m9" },
    { 5, {{1, 0}, {9, -1}, {3, -1}, {5,  0}, {7, -1}}, "m♭9" },
    { 5, {{1, 0}, {9,  0}, {3,  0}, {5, -1}, {7,  0}}, "M9" SUP "♭5" PUS },
    { 5, {{1, 0}, {9,  0}, {3,  0}, {5, -1}, {7, -1}}, "9" SUP "♭5" PUS },
    { 5, {{1, 0}, {9,  0}, {3, -1}, {5, -1}, {7,  0}}, "m" SUP "M9♭5" PUS },
    { 5, {{1, 0}, {9,  0}, {3, -1}, {5, -1}, {7, -1}}, "𝆩" SUP "9" PUS },
    { 5, {{1, 0}, {9, -1}, {3, -1}, {5, -1}, {7, -1}}, "𝆩" SUP "♭9" PUS },
    { 5, {{1, 0}, {9,  0}, {3, -1}, {5, -1}, {7, -2}}, "°" SUP "9" PUS },
    { 5, {{1, 0}, {9, -1}, {3, -1}, {5, -1}, {7, -2}}, "°" SUP "♭9" PUS },

    /* Eleventh chords */
    { 6, {{1, 0}, {9,  0}, {3,  0}, {11,  0}, {5, +1}, {7,  0}}, SUP "+M11" PUS },
    { 6, {{1, 0}, {9,  0}, {3,  0}, {11,  0}, {5, +1}, {7, -1}}, SUP "+11" PUS },
    { 6, {{1, 0}, {9,  0}, {3,  0}, {11,  0}, {5,  0}, {7,  0}}, "M11" },
    { 6, {{1, 0}, {9,  0}, {3,  0}, {11,  0}, {5,  0}, {7, -1}}, "11" },
    { 6, {{1, 0}, {9,  0}, {3, -1}, {11,  0}, {5,  0}, {7,  0}}, "m" SUP "M11" PUS },
    { 6, {{1, 0}, {9,  0}, {3, -1}, {11,  0}, {5,  0}, {7, -1}}, "m11" },
    { 6, {{1, 0}, {9, -1}, {3, -1}, {11,  0}, {5, -1}, {7, -1}}, "𝆩" SUP "11" PUS },
    { 6, {{1, 0}, {9, -1}, {3, -1}, {11, -1}, {5, -1}, {7, -2}}, "°" SUP "11" PUS }
};

static void
scherzo_adjust_pitch_to_match_modified_degree (pitch_t *pitch,
					       pitch_t root,
					       modified_degree_t *degree)
{
    pitch_name_t name = PITCH_NAME (*pitch);
    pitch_accidental_t accidental = PITCH_ACCIDENTAL (*pitch);
    int octave = PITCH_OCTAVE (*pitch);
    int degree_zero_based = (degree->degree - 1) % 7;
    int degree_delta;

    int note_degree_zero_based = name - PITCH_NAME (root);
    if (note_degree_zero_based < 0)
	note_degree_zero_based += 7;

    if (note_degree_zero_based == degree_zero_based)
	return;

    degree_delta = note_degree_zero_based - degree_zero_based;
    if (degree_delta > 3)
	degree_delta = - (7 - degree_delta);
    if (degree_delta < -3)
	degree_delta = - (-7 - degree_delta);

    if (abs (degree_delta) != 1) {
	fprintf (stderr, "Internal error: Cannot adjust a note more than one degree (%d vs. %d).\n", note_degree_zero_based + 1, degree->degree);
	exit (1);
    }

    if (degree_delta == -1) {
	if (name == PITCH_NAME_B) {
	    name = PITCH_NAME_C;
	    octave++;
	} else {
	    name++;
	}
	switch (name) {
	case PITCH_NAME_D:
	case PITCH_NAME_E:
	case PITCH_NAME_G:
	case PITCH_NAME_A:
	case PITCH_NAME_B:
	    accidental -= 2;
	    break;
	case PITCH_NAME_C:
	case PITCH_NAME_F:
	    accidental -= 1;
	    break;
	}
    }

    if (degree_delta == +1) {
	if (name == PITCH_NAME_C) {
	    name = PITCH_NAME_B;
	    octave--;
	} else {
	    name--;
	}
	switch (name) {
	case PITCH_NAME_C:
	case PITCH_NAME_D:
	case PITCH_NAME_F:
	case PITCH_NAME_G:
	case PITCH_NAME_A:
	    accidental += 2;
	    break;
	case PITCH_NAME_E:
	case PITCH_NAME_B:
	    accidental += 1;
	}
    }

    if (accidental < 0 || accidental > PITCH_ACCIDENTAL_DOUBLE_SHARP) {
	fprintf (stderr, "Internal error: Trying to adjust an accidental out of range (degree_delta == %d (%d - %d), new accidental == %d).\n", degree_delta, note_degree_zero_based, degree_zero_based, accidental);
	exit (1);
    }

    *pitch = PITCH (name, accidental, octave);
}

static void
_spell_chord_by_signature (pitch_group_t *chord,
			   int num_notes,
			   chord_signature_t *signature,
			   pitch_t root)
{
    int i, degree, note_half_steps, degree_half_steps;
    int root_midi;

    /* Sanitize root to eliminate things like double-flats,
     * double-sharps, Cb, E#, etc. */
    root_midi = _pitch_to_midi (root);
    root = _midi_to_pitch (root_midi);
    
    for (i = 0; i < num_notes; i++) {
	note_half_steps = _pitch_from_root_in_half_steps (chord->pitches[i],
							  root);
	for (degree = 0; degree < signature->num_degrees; degree++) {
	    degree_half_steps = _modified_degree_to_half_steps (&signature->degrees[degree]);
	    if (note_half_steps == degree_half_steps) {
		scherzo_adjust_pitch_to_match_modified_degree (&chord->pitches[i],
							       root,
							       &signature->degrees[degree]);
		break;
	    }
	}
	if (note_half_steps != degree_half_steps) {
	    fprintf (stderr, "Internal error: Chord and degree mis-match\n");
	    exit (1);
	}
    }
}

static void
scherzo_analyze_chord (scherzo_t *scherzo)
{
    void *local = talloc_new (NULL);
    analyzed_pitch_t *pitches;
    pitch_group_t *pitch_group;
    unsigned i, j, num_pitches;
    int bass_pitch, inversion;
    pitch_t root;
    chord_signature_t *match = NULL;
    char *chord_name;

    if (scherzo->pedal_pressed)
	pitch_group = &scherzo->notes_pedaled;
    else
	pitch_group = &scherzo->notes_pressed;

    num_pitches = pitch_group->num_pitches;

    if (scherzo->chord) {
	score_remove_chord (scherzo->chord);
	scherzo->chord = NULL;
    }

    if (num_pitches <= 1)
	goto DONE;

    pitches = talloc_array (local, analyzed_pitch_t, num_pitches);
    if (pitches == NULL)
	goto DONE;

    for (i = 0; i < num_pitches; i++) {
	pitch_t pitch = pitch_group->pitches[i];
	pitches[i].pitch = pitch;
	pitches[i].midi_pitch = _pitch_to_midi (pitch);
	/* Relative pitch will be filled in after sorting. */
	pitches[i].relative_pitch = 0;
    }

    /* First, sort by midi pitch to find the bass note. */
    qsort (pitches, num_pitches, sizeof (analyzed_pitch_t),
	   _compare_analyzed_pitch_by_midi_pitch);
    
    bass_pitch = pitches[0].midi_pitch;

    /* With the bass note identified, we can find all relative pitches. */
    for (i = 0; i < num_pitches; i++) {
	pitches[i].relative_pitch = pitches[i].midi_pitch - bass_pitch;
	while (pitches[i].relative_pitch >= 12)
	    pitches[i].relative_pitch -= 12;
    }

    /* Now, sort again by relative pitch. */
    qsort (pitches, num_pitches, sizeof (analyzed_pitch_t),
	   _compare_analyzed_pitch_by_relative_pitch);

    /* Finally, eliminate all duplicate notes. */
    for (i = 0; i < num_pitches - 1; i++) {
	    if (pitches[i+1].relative_pitch == pitches[i].relative_pitch) {
		    j = i+1;
		    while (j < num_pitches &&
			   pitches[j].relative_pitch == pitches[i].relative_pitch)
		    {
			    j++;
		    }
		    /* The loop incremented j one past the last
		     * duplicate. Decrement so that it points to the
		     * last duplicate (and is guaranteed to not exceed
		     * the array bounds).*/
		    j--;

		    if (j < num_pitches - 1) {
			    memmove (&pitches[i+1], &pitches[j+1],
				     (num_pitches - j) * sizeof (analyzed_pitch_t));
		    }

		    num_pitches -= (j - i);
	    }
    }

    for (inversion = 0; inversion < 4; inversion++) {
	for (i = 0; i < ARRAY_SIZE (signatures); i++) {
	    if (_chord_signature_matches (pitches, num_pitches,
					  signatures[i].degrees,
					  signatures[i].num_degrees,
					  inversion, &root))
	    {
		match = &signatures[i];
		goto HAVE_MATCH;
	    }
	}
    }
HAVE_MATCH:

    chord_name = (char *) signatures[i].name;

    if (match) {
	/* Don't print root pitch for octaves and inversions,
	 * (since a pitch name alone looks like a major triad) */
	if (num_pitches < 3) {
	    chord_name = talloc_strdup (local, chord_name);
	} else {
	    chord_name = talloc_asprintf (local, "%s%s",
					  _pitch_str (root),
					  chord_name);
	}

	_spell_chord_by_signature (pitch_group, num_pitches,
				   match, root);
    } else {
	chord_name = talloc_strdup (local, "?");
    }

    scherzo->chord = score_add_chord (scherzo->treble, chord_name);

DONE:
    talloc_free (local);
}

static void
pitch_group_init (void *ctx, pitch_group_t *group)
{
    group->ctx = ctx;
    group->pitches = NULL;
    group->size = 0;
    group->num_pitches = 0;
}

static void
pitch_group_add_pitch (pitch_group_t *group, pitch_t pitch)
{
    int i;

    /* Do nothing if pitch is already in group. */
    for (i = 0; i < group->num_pitches; i++)
	if (group->pitches[i] == pitch)
	    return;

    group->num_pitches++;

    if (group->num_pitches > group->size) {
	group->size++;
	group->pitches = talloc_realloc (group->ctx, group->pitches,
					 pitch_t, group->size);

	if (group->pitches == NULL) {
	    fprintf (stderr, "Out of memory.\n");
	    exit (1);
	}
    }

    group->pitches[group->num_pitches - 1] = pitch;
}

static void
pitch_group_remove_pitch_at (pitch_group_t *group, int i)
{
    if (i > group->num_pitches) {
	fprintf (stderr, "Internal error: Cannot remove pitch %d from group with only %d %s\n",
		 i, group->num_pitches,
		 group->num_pitches == 1 ? "pitch" : "pitches");
	exit (1);
    }

    if (i < group->num_pitches - 1) {
	memmove (group->pitches + i, group->pitches + i + 1,
		 (group->num_pitches - 1 - i) * sizeof (pitch_t));
    }

    group->num_pitches--;
}

static void
pitch_group_remove_pitch (pitch_group_t *group, pitch_t pitch)
{
    int i;

    for (i = 0; i < group->num_pitches; i++)
	if (group->pitches[i] == pitch)
	    pitch_group_remove_pitch_at (group, i);
}

static void
scherzo_update_notes_and_chord (scherzo_t *scherzo)
{
    if (scherzo->notes_pressed.num_pitches == 0 &&
	scherzo->notes_pedaled.num_pitches == 0)
    {
	score_remove_notes (scherzo->score);

	if (scherzo->chord)
	    score_remove_chord (scherzo->chord);

	gtk_widget_queue_draw (scherzo->window);
    }
}

static void
scherzo_press_note (scherzo_t *scherzo, pitch_t pitch)
{
    int i;

    /* Do nothing if this note is already pressed. */
    for (i = 0; i < scherzo->notes_pressed.num_pitches; i++)
	if (scherzo->notes_pressed.pitches[i] == pitch)
	    return;

    pitch_group_add_pitch (&scherzo->notes_pressed, pitch);

    if (scherzo->pedal_pressed)
	pitch_group_add_pitch (&scherzo->notes_pedaled, pitch);

    /* Remove all notes from score, then add current notes. */
    score_remove_notes (scherzo->score);

    for (i = 0; i < scherzo->notes_pressed.num_pitches; i++)
    {
	score_add_note (scherzo->score, scherzo->notes_pressed.pitches[i],
			SCORE_DURATION_WHOLE);
    }

    for (i = 0; i < scherzo->notes_pedaled.num_pitches; i++)
    {
	score_add_note (scherzo->score, scherzo->notes_pedaled.pitches[i],
			SCORE_DURATION_WHOLE);
    }

    scherzo_analyze_chord (scherzo);
}

static void
scherzo_release_note (scherzo_t *scherzo, pitch_t pitch)
{
    pitch_group_remove_pitch (&scherzo->notes_pressed, pitch);

    if (scherzo->notes_pressed.num_pitches == 0)
	scherzo_update_notes_and_chord (scherzo);
}

static pitch_t
scherzo_press_note_midi (scherzo_t *scherzo, unsigned char midi_note)
{
    pitch_t pitch = _midi_to_pitch (midi_note);

    scherzo_press_note (scherzo, pitch);

    return pitch;
}

static void
scherzo_release_note_midi (scherzo_t *scherzo, unsigned char midi_note)
{
    scherzo_release_note (scherzo, _midi_to_pitch (midi_note));
}

static void
scherzo_press_pedal (scherzo_t *scherzo)
{
    int i;

    scherzo->pedal_pressed = 1;

    /* Copy all pressed notes to pedaled notes */
    for (i = 0; i < scherzo->notes_pressed.num_pitches; i++)
	pitch_group_add_pitch (&scherzo->notes_pedaled, scherzo->notes_pressed.pitches[i]);
}

static void
scherzo_release_pedal (scherzo_t *scherzo)
{
    int i;

    /* Remove all previously pedaled notes. */
    for (i = scherzo->notes_pedaled.num_pitches - 1; i >=0; i--)
	pitch_group_remove_pitch_at (&scherzo->notes_pedaled, i);

    scherzo->pedal_pressed = 0;

    scherzo_update_notes_and_chord (scherzo);
}

void
_select_challenge (scherzo_t *scherzo)
{
    category_t *category_unused;
    bool_t introduced_unused;
    item_t *item;
    challenge_t *challenge = &scherzo->challenge;
    pitch_name_t pitch_name;
    int octave;
    char *s;

    mnemon_select_item (&scherzo->mnemon,
			&challenge->bin,
			&challenge->item_index,
			&category_unused,
			&introduced_unused);

    item = challenge->bin->items[challenge->item_index];

    s = item->challenge;
    if (strncmp (s, "treble:", 7) == 0) {
	s += 7;
	challenge->staff = scherzo->treble;
    } else if (strncmp (s, "bass:", 5) == 0) {
	s += 5;
	challenge->staff = scherzo->bass;
    } else {
	fprintf (stderr,
		 "Malformed staff name: %s (expected 'treble:' or 'bass:')\n",
		 s);
	exit (1);
    }

    switch (*s) {
    case 'C':
	pitch_name = PITCH_NAME_C;
	break;
    case 'D':
	pitch_name = PITCH_NAME_D;
	break;
    case 'E':
	pitch_name = PITCH_NAME_E;
	break;
    case 'F':
	pitch_name = PITCH_NAME_F;
	break;
    case 'G':
	pitch_name = PITCH_NAME_G;
	break;
    case 'A':
	pitch_name = PITCH_NAME_A;
	break;
    case 'B':
	pitch_name = PITCH_NAME_B;
	break;
    default:
	fprintf (stderr, "Malformed pitch name: %s (expected 'A' - 'G')\n", s);
	exit (1);
    }
    s++;

    if (*s < '0' || *s > '9') {
	fprintf (stderr, "Malformed octave number: %s (expected '0' - '9')\n", s);
	exit (1);
    }

    octave = *s - '0';

    challenge->pitch = PITCH (pitch_name, PITCH_ACCIDENTAL_NATURAL, octave);

    challenge->active = 1;
    challenge->satisfied = 0;
    challenge->mistaken = 0;
}

/* Determine whether the user hit the correct note. */
static void
_judge_pitch (scherzo_t *scherzo, pitch_t pitch)
{
    challenge_t *challenge = &scherzo->challenge;

    if (! challenge->active)
	return;

    if (pitch == challenge->pitch)
	challenge->satisfied = 1;
    else
	challenge->mistaken = 1;
}

/* If the user got the right note (eventually), then score it in
 * mnemon and show the next note. */
static void
_score_challenge (scherzo_t *scherzo)
{
    challenge_t *challenge = &scherzo->challenge;

    if (! challenge->active)
	return;

    if (! challenge->satisfied)
	return;

    mnemon_score_item (&scherzo->mnemon, challenge->bin, challenge->item_index,
		       ! challenge->mistaken);

    _select_challenge (scherzo);
}

static int
on_midi_input (unused (GIOChannel *channel),
	       unused (GIOCondition condition),
	       void *user_data)
{
    unsigned char buf[MIDI_BUF_SIZE], *next;
    scherzo_t *scherzo = user_data;
    ssize_t remaining;
    snd_seq_event_t event;
    pitch_t pitch;
    int need_redraw = FALSE;

    remaining = read (scherzo->midi_fd, buf, MIDI_BUF_SIZE);

    next = buf;
    while (remaining) {
	long consumed;

	consumed = snd_midi_event_encode (scherzo->snd_midi_event,
					  next, remaining, &event);

	remaining -= consumed;
	next += consumed;

	switch (event.type) {
	case SND_SEQ_EVENT_NONE:
	    /* Incomplete event. Nothing to do. */
	    break;
	case SND_SEQ_EVENT_NOTEON:
	    pitch = scherzo_press_note_midi (scherzo, event.data.note.note);
	    _judge_pitch (scherzo, pitch);
	    need_redraw = TRUE;
	    break;
	case SND_SEQ_EVENT_NOTEOFF:
	    scherzo_release_note_midi (scherzo, event.data.note.note);
	    _score_challenge (scherzo);
	    need_redraw = TRUE;
	    break;
	case SND_SEQ_EVENT_CLOCK:
	    /* Ignore for now as my piano sends a constant stream of these. */
	    break;
	case SND_SEQ_EVENT_SENSING:
	    /* Ignore for now as my piano sends a constant stream of these. */
	    break;
	case SND_SEQ_EVENT_CONTROLLER:
	    /* XXX: My piano gives me 64 for the sustain pedal, is
	     * that universal? */
	    if (event.data.control.param == 64) {
		if (event.data.control.value == 0)
		    scherzo_release_pedal (scherzo);
		else
		    scherzo_press_pedal (scherzo);
	    } else {
		fprintf (stderr, "Fixme: Unhandled MIDI Control event (param=%d, value=%d)\n",
			 event.data.control.param, event.data.control.value);
	    }
	    break;
	default:
	    fprintf (stderr, "Fixme: Do not yet know how to handle MIDI event %d\n",
		     event.type);
	    break;
	}
    }

    if (need_redraw)
	gtk_widget_queue_draw (scherzo->window);

    /* Return TRUE to continue to get called in the future. */
    return TRUE;
}

int
main (int argc, char *argv[])
{
    GtkWidget *drawing_area;
    scherzo_t scherzo;
    int err;

    srand (time (NULL));

    gtk_init (&argc, &argv);

    scherzo.ctx = talloc_new (NULL);

    scherzo.score = score_create (scherzo.ctx);
    scherzo.staff_height = 100;
    score_set_staff_height (scherzo.score, scherzo.staff_height);

    score_add_brace (scherzo.score, 2);
    scherzo.treble = score_add_staff (scherzo.score, SCORE_CLEF_G);
    scherzo.bass = score_add_staff (scherzo.score, SCORE_CLEF_F);

    scherzo.chord = NULL;

    /* Default to octave 4 and natural for computer keyboard keypresses. */
    scherzo.keyboard_octave = 4;
    scherzo.keyboard_accidental = PITCH_ACCIDENTAL_NATURAL;

    pitch_group_init (scherzo.ctx, &scherzo.notes_pressed);
    pitch_group_init (scherzo.ctx, &scherzo.notes_pedaled);

    scherzo.pedal_pressed = 0;

    mnemon_init (&scherzo.mnemon);
    /* XXX: Should create a default file if one cannot be loaded. */
    mnemon_load_category (&scherzo.mnemon, "scherzo-notes");

    scherzo.challenge.active = 0;
/*
    _select_challenge (&scherzo);
*/

    err = snd_midi_event_new (MIDI_BUF_SIZE, &scherzo.snd_midi_event);
    if (err) {
	fprintf (stderr, "Out of memory.\n");
	return 1;
    }

#define MIDI_DEVICE "/dev/midi1"
    scherzo.midi_fd = open (MIDI_DEVICE, O_RDONLY);
    if (scherzo.midi_fd < 0) {
	printf ("failed to open " MIDI_DEVICE ". Midi input will not be available.\n");
    } else {
	GIOChannel *channel;

	channel = g_io_channel_unix_new (scherzo.midi_fd);
	g_io_channel_set_encoding (channel, NULL, NULL);
	g_io_add_watch (channel, G_IO_IN, on_midi_input, &scherzo);
    }

    scherzo.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

    gtk_window_set_default_size (GTK_WINDOW (scherzo.window), 1000, 600);

    g_signal_connect (scherzo.window, "delete-event",
		      G_CALLBACK (on_delete_event_quit), NULL);

    drawing_area = gtk_drawing_area_new ();

    gtk_container_add (GTK_CONTAINER (scherzo.window), drawing_area);

    g_signal_connect (drawing_area, "expose-event",  
		      G_CALLBACK (on_expose_event_draw),
		      &scherzo);

    g_signal_connect (scherzo.window, "key-press-event",
		      G_CALLBACK (on_key_press_event),
		      &scherzo);

    g_signal_connect (scherzo.window, "key-release-event",
		      G_CALLBACK (on_key_release_event),
		      &scherzo);
    
    gtk_widget_show_all (scherzo.window);
    
    gtk_main ();

    mnemon_save (&scherzo.mnemon);

    snd_midi_event_free (scherzo.snd_midi_event);

    talloc_free (scherzo.ctx);

    return 0;
}
